@@ -15,6 +15,7 @@
 //! Homepage: <https://github.com/gwbres/cggtts>
 
 pub mod track;
+use std::io::{Write};
 use regex::Regex;
 use thiserror::Error;
 use std::str::FromStr;
@@ -25,7 +26,7 @@ use scan_fmt::scan_fmt;
 const VERSION: &str = "2E";
 
 /// latest revision date
-const REV_DATE: &str = "2014-02-20";
+const LATEST_REV_DATE: &str = "2014-02-20";
 
 #[derive(Clone, Debug)]
 /// `Rcvr` describes a GNSS receiver
@@ -39,6 +40,21 @@ pub struct Rcvr {
     software_number: String,
 }
 
+impl std::fmt::Display for Rcvr { 
+    fn fmt(&self, fmt: &mut std::fmt::Formatter) -> std::fmt::Result {
+        fmt.write_str(&self.manufacturer)?;
+        fmt.write_str(" ")?;
+        fmt.write_str(&self.recv_type)?;
+        fmt.write_str(" ")?;
+        fmt.write_str(&self.serial_number)?;
+        fmt.write_str(" ")?;
+        fmt.write_str(&self.year.to_string())?;
+        fmt.write_str(" ")?;
+        fmt.write_str(&self.software_number)?;
+        Ok(())
+    }
+}
+
 #[derive(Clone, Debug)]
 /// `CalibratedDelay` are delays that are
 /// specified to a specific carrier frequency,
@@ -63,6 +79,25 @@ impl Default for CalibratedDelay {
     }
 }
 
+impl std::fmt::Display for CalibratedDelay { 
+    fn fmt(&self, fmt: &mut std::fmt::Formatter) -> std::fmt::Result {
+        if self.values.len() == 1 {
+            fmt.write_str(&format!("{:.1} ns ({} {})", self.values[0] * 1E9, self.constellation, self.codes[0]))?
+        } else {
+            // CSV
+            for i in 0..self.values.len()-1 {
+                fmt.write_str(&format!("{:.1} ns ({} {}), ", 
+                    self.values[i] *1E9, self.constellation, self.codes[i]))? 
+            }
+            fmt.write_str(&format!("{:.1} ns ({} {})", 
+                self.values[self.values.len()-1] *1E9, self.constellation, 
+                    self.codes[self.values.len()-1]))?
+        }
+        fmt.write_str(&format!("     CAL_ID = {}", self.report))?;
+        Ok(())
+    }
+}
+
 impl CalibratedDelay {
     /// Builds a new `CalibratedDelay` object
     pub fn new(constellation: track::Constellation, values: Vec<f64>, codes: Vec<String>, report: Option<&str>) -> CalibratedDelay {
@@ -199,7 +234,7 @@ impl Default for Cggtts {
     fn default() -> Cggtts {
         Cggtts {
             version: VERSION.to_string(),
-            rev_date: chrono::NaiveDate::parse_from_str(REV_DATE, "%Y-%m-%d")
+            rev_date: chrono::NaiveDate::parse_from_str(LATEST_REV_DATE, "%Y-%m-%d")
                 .unwrap(),
             date: chrono::Utc::today().naive_utc(),
             lab: String::from("Unknown"),
@@ -250,6 +285,11 @@ impl Cggtts {
     /// Returns GNSS receiver number of channels
     pub fn get_nb_channels (&self) -> u16 { self.nb_channels }
 
+    /// Assigns `Rcvr` hardware description
+    pub fn set_rcvr_infos (&mut self, rcvr: Rcvr) { self.rcvr = Some(rcvr) }
+    /// Assigns `IMS` evluation device description
+    pub fn set_ims_infos (&mut self, ims: Rcvr) { self.ims = Some(ims) }
+
     /// Assigns antenna phase center coordinates [m],
     /// coordinates should use `IRTF` referencing
     pub fn set_antenna_coordinates (&mut self, coords: (f32,f32,f32)) { self.coordinates = coords }
@@ -309,7 +349,10 @@ impl Cggtts {
                                     ret.values.push(delay.values[i]);
                                 }
                             },
-                            None => {} // no delay at all, 0 assumed then
+                            None => { // no delay at all, 0 assumed then
+                                ret.values.push(0.0);
+                                ret.codes.push(String::from("C1"))
+                            },
                         }
                     }
                 }
@@ -336,8 +379,10 @@ impl Cggtts {
     /// Appends one track to self (if possible)
     pub fn push_track (&mut self, track: track::CggttsTrack) { self.tracks.push(track) }
 
-    /// returns true if self is `Dual Frequency Cggtts`
-    pub fn is_dual_frequency (&self) -> bool { self.total_delay().values.len() == 1 }
+    /// returns true if self is `Single Frequency Cggtts`
+    pub fn is_single_frequency (&self) -> bool { self.total_delay().values.len() == 1 }
+    /// returns true if self is `Single Frequency Cggtts`
+    pub fn is_dual_frequency (&self) -> bool { !self.is_single_frequency() }
 
     /// Returns true if self contains ionospheric information
     pub fn has_ionospheric_parameters (&self) -> bool {
@@ -363,8 +408,7 @@ impl Cggtts {
     /// Sets `ref` delay (refer to README)
     pub fn set_ref_delay (&mut self, delay: f64) { self.ref_dly = delay }
 
-    /// Builds `Cggtts` from given file.
-    /// File must respect naming convention.
+    /// Builds self from given `Cggtts` file.
     pub fn from_file (fp: &std::path::Path) -> Result<Cggtts, Error> {
         let file_name = fp.file_name()
             .unwrap()
@@ -575,7 +619,6 @@ impl Cggtts {
                 cleanedup = before.strip_suffix(" CAL_ID =")
                     .unwrap()
                     .trim();
-                println!("CLEANED UP '{}'", cleanedup);
                 // final delay identification
                 let (constellation, values, codes) : 
                     (track::Constellation, Vec<f64>, Vec<String>)
@@ -682,27 +725,88 @@ impl Cggtts {
             tracks
         })
     }
-}
-
-// custom display formatter
-impl std::fmt::Display for Cggtts {
-    fn fmt (&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        write!(
-            f, "Version: '{}' | REV DATE '{}' | LAB '{}' | Nb Channels {}\nRECVR: {:?}\nIMS: {:?}\nCoordinates: {:?}\nFRAME: {}\nCOMMENTS: {:#?}\nREFERENCE: {}\n",
-            self.version, self.rev_date, self.lab, self.nb_channels,
-            self.rcvr,
-            self.ims,
-            self.coordinates,
-            self.frame,
-            self.comments,
-            self.reference,
-        ).unwrap();
-        write! (f, "-------------------------\n").unwrap();
-        for i in 0..self.tracks.len() {
-            write! (f, "MEAS #{}: {}\n",i, self.tracks[i]).unwrap()
+    
+    /// Writes self into a `Cggtts` file
+    pub fn to_file (&self, fp: &std::path::Path) -> std::io::Result<()> {
+        let mut fd = std::fs::File::create(&fp)?;
+        writeln!(&mut fd,
+"CGGTTS GENERIC DATA FORMAT VERSION = {}
+REV DATE = {}
+RCVR = {}
+CH = {}
+IMS = {}
+LAB = {}
+X = {}
+Y = {}
+Z = {}
+FRAME = {}
+COMMENTS = {}",
+            VERSION,
+            LATEST_REV_DATE,
+            self.rcvr.clone()
+                .map(|r| r.to_string())
+                .unwrap_or(String::from("RRRRRRRR")),
+            self.nb_channels,
+            self.ims.clone()
+                .map(|r| r.to_string())
+                .unwrap_or(String::from("99999")),
+            self.lab,
+            format!("{:.3}", self.coordinates.0),
+            format!("{:.3}", self.coordinates.1),
+            format!("{:.3}", self.coordinates.2),
+            &self.frame,
+            self.comments.clone()
+                .map(|r| r.to_string())
+                .unwrap_or(String::from("NO COMMENTS")),
+        )?;
+        // system delays
+        if let Some(delay) = &self.tot_dly {
+            // total delay defined
+            writeln!(&mut fd, "TOT DLY = {}", delay.to_string())?;
+        } else {
+            // total delay not defined
+            // => SYS or INT DELAY ?
+            // INT DELAY prioritary
+            if let Some(delay) = &self.int_dly {
+                writeln!(&mut fd, "INT DLY = {}", delay.to_string())?;
+            } else if let Some(delay) = &self.sys_dly {
+                writeln!(&mut fd, "SYS DLY = {}", delay.to_string())?;
+            } else {
+                // neither SYS / INT delay
+                // => specify null SYS DLY
+                let null_delay = CalibratedDelay {
+                    constellation: track::Constellation::default(),
+                    values: vec![0.0_f64],
+                    codes: vec![String::from("C1")],
+                    report: String::from("NA"),
+                };
+                writeln!(&mut fd, "SYS DLY = {}", null_delay.to_string())? 
+            }
+            // other delays always there
+            writeln!(&mut fd, "CAB DLY = {} ns", format!("{:.1}", self.cab_dly *1E9))?;
+            writeln!(&mut fd, "REF DLY = {} ns", format!("{:.1}", self.ref_dly *1E9))?;
         }
-        write!(f, "\n")
+        writeln!(&mut fd, "REF = {}", self.reference.to_string())?;
+        // <!> TODO <!>
+        // CKSUM = {:2X}
+
+        writeln!(&mut fd)?; // blank separator
+        if self.has_ionospheric_parameters() {
+            writeln!(&mut fd, "{}", track::TRACK_LABELS_WITH_IONOSPHERIC_DATA)?;
+            writeln!(&mut fd, "{}",
+"              hhmmss s .1dg .1dg .1ns .1ps/s .1ns .1ps/s .1ns .1ns.1ps/s.1ns.1ps/s.1ns.1ps/s.1ns")?
+        } else {
+            writeln!(&mut fd, "{}", track::TRACK_LABELS_WITHOUT_IONOSPHERIC_DATA)?;
+            writeln!(&mut fd, "{}",
+"             hhmmss s   .1dg .1dg    .1ns     .1ps/s     .1ns    .1ps/s .1ns     .1ns.1ps/s.1ns.1ps/s")?
+        }
+
+        for i in 0..self.tracks.len() {
+            writeln!(&mut fd, "{}", self.tracks[i])?
+        }
+        Ok(())
     }
+
 }
 
 #[cfg(test)]
@@ -713,21 +817,23 @@ mod test {
     /// Tests default constructor 
     fn cggtts_test_default() {
         let cggtts = Cggtts::new();
-        let today = chrono::Utc::today();
-        let rev_date = chrono::NaiveDate::parse_from_str(REV_DATE,"%Y-%m-%d")
-            .unwrap();
-        assert_eq!(cggtts.lab, "Unknown");
-        assert_eq!(cggtts.nb_channels, 0);
-        assert_eq!(cggtts.frame, "?");
-        assert_eq!(cggtts.reference, "Unknown");
-        //assert_eq!(cggtts.tot_dly, None);
-        //assert_eq!(cggtts.ref_dly, None);
-        //assert_eq!(cggtts.int_dly, None);
-        //assert_eq!(cggtts.cab_dly, None);
-        //assert_eq!(cggtts.sys_dly, None);
-        assert_eq!(cggtts.coordinates, (0.0,0.0,0.0));
-        assert_eq!(cggtts.rev_date, rev_date); 
-        assert_eq!(cggtts.date, today.naive_utc());
+        assert_eq!(cggtts.lab, "Unknown"); // default
+        assert_eq!(cggtts.nb_channels, 0); // default
+        assert_eq!(cggtts.frame, "?"); // default ..
+        assert_eq!(cggtts.reference, "Unknown"); // default..
+        assert_eq!(cggtts.coordinates, (0.0,0.0,0.0)); // empty..
+        assert_eq!(cggtts.rev_date,
+            chrono::NaiveDate::parse_from_str(LATEST_REV_DATE, "%Y-%m-%d")
+            .unwrap());
+        assert_eq!(cggtts.date, chrono::Utc::today().naive_utc());
+        assert_eq!(cggtts.tot_dly.is_none(), true);
+        assert_eq!(cggtts.int_dly.is_none(), true);
+        assert_eq!(cggtts.sys_dly.is_none(), true);
+        assert_eq!(cggtts.cab_dly, 0.0);
+        assert_eq!(cggtts.ref_dly, 0.0);
+        println!("{:#?}", cggtts.total_delay());
+        assert_eq!(cggtts.total_delay().values.len(), 1); // single freq Cggts by default
+        assert_eq!(cggtts.total_delay().values[0], 0.0); // not specified
         println!("{:#?}", cggtts)
     }
 
@@ -801,6 +907,7 @@ mod test {
     
     #[test]
     /// Tests standard file parsing
+/*
     fn cggtts_test_from_standard_data() {
         // open test resources
         let test_resources = std::path::PathBuf::from(
@@ -824,7 +931,7 @@ mod test {
             }
         }
     }
-    
+*/
     #[test]
     /// Tests advanced file parsing
     fn cggtts_test_from_ionospheric_data() {
@@ -850,4 +957,217 @@ mod test {
             }
         }
     }
+
+    #[test]
+    /// Tests basci `Cggtts` to file
+    fn default_cggtts_to_file() {
+        let cggtts = Cggtts::default();
+        let fp = std::path::Path::new("data/output/GZXXXXDD.DD0");
+        assert_eq!(cggtts.to_file(fp).is_err(), false)
+    }
+
+    #[test]
+    /// Tests customized `Cggtts` to file
+    fn basic_cggtts_to_file() {
+        let mut cggtts = Cggtts::default();
+
+        // identify receiver hw
+        let rcvr = Rcvr {
+            manufacturer: String::from("SomeManufacturer"),
+            recv_type: String::from("SomeKind"), 
+            serial_number: String::from("XXXXXX"), 
+            year: 2021, 
+            software_number: String::from("v00"),
+        };
+        cggtts.set_rcvr_infos(rcvr);
+
+        // add some more infos
+        cggtts.set_lab_agency("MyLab");
+        cggtts.set_nb_channels(10);
+        cggtts.set_antenna_coordinates((1.0,2.0,3.0));
+        cggtts.set_time_reference("UTC(k)");
+
+        // define a total delay
+        let delay = CalibratedDelay {
+            constellation: track::Constellation::Glonass,
+            values: vec![100E-9_f64],
+            codes: vec![String::from("C1")], 
+            report: String::from("NA"),
+        };
+        cggtts.set_total_delay(delay);
+
+        let fp = std::path::Path::new("data/output/GZXXXXDD.DD1");
+        assert_eq!(cggtts.to_file(fp).is_err(), false)
+    }
+    
+    #[test]
+    /// Tests customized `Cggtts` to file
+    fn dual_frequency_cggtts_to_file() {
+        let mut cggtts = Cggtts::default();
+
+        // identify receiver hw
+        let rcvr = Rcvr {
+            manufacturer: String::from("SomeManufacturer"),
+            recv_type: String::from("SomeKind"), 
+            serial_number: String::from("XXXXXX"), 
+            year: 2021, 
+            software_number: String::from("v00"),
+        };
+        cggtts.set_rcvr_infos(rcvr);
+
+        // add some more infos
+        cggtts.set_lab_agency("MyLab");
+        cggtts.set_nb_channels(10);
+        cggtts.set_antenna_coordinates((1.0,2.0,3.0));
+        cggtts.set_time_reference("UTC(k)");
+
+        // set a total delay
+        let total_delay = CalibratedDelay {
+            constellation: track::Constellation::Glonass,
+            values: vec![100E-9, 150E-9],
+            codes: vec![String::from("C1"),String::from("C2")], 
+            report: String::from("NA"),
+        };
+        cggtts.set_total_delay(total_delay);
+        println!("{:#?}",cggtts);
+
+        let fp = std::path::Path::new("data/output/GZXXXXDD.DD2");
+        assert_eq!(cggtts.to_file(fp).is_err(), false)
+    }
+    
+    #[test]
+    /// Tests customized `Cggtts` to file (B)
+    fn cggtts_with_system_delay_to_file() {
+        let mut cggtts = Cggtts::default();
+
+        // identify receiver hw
+        let rcvr = Rcvr {
+            manufacturer: String::from("SomeManufacturer"),
+            recv_type: String::from("SomeKind"), 
+            serial_number: String::from("XXXXXX"), 
+            year: 2021, 
+            software_number: String::from("v00"),
+        };
+        cggtts.set_rcvr_infos(rcvr);
+
+        // add some more infos
+        cggtts.set_lab_agency("MyLab");
+        cggtts.set_nb_channels(10);
+        cggtts.set_antenna_coordinates((1.0,2.0,3.0));
+        cggtts.set_time_reference("UTC(k)");
+
+        // define a total delay
+        let delay = CalibratedDelay {
+            constellation: track::Constellation::Glonass,
+            values: vec![100E-9_f64],
+            codes: vec![String::from("C2")], 
+            report: String::from("NA"),
+        };
+        cggtts.set_system_delay(delay);
+        cggtts.set_cable_delay(50E-9);
+        cggtts.set_ref_delay(100E-9);
+        let total_delay = cggtts.total_delay();
+        assert_eq!(total_delay.values.len(), 1); // single freq
+        assert_eq!(total_delay.values[0], 100E-9+50E-9); // single freq
+
+        let fp = std::path::Path::new("data/output/GZXXXXDD.DD3");
+        assert_eq!(cggtts.to_file(fp).is_err(), false)
+    }
+    
+    #[test]
+    /// Tests customized `Cggtts` to file (C)
+    fn cggtts_with_internal_delay_to_file() {
+        let mut cggtts = Cggtts::default();
+
+        // identify receiver hw
+        let rcvr = Rcvr {
+            manufacturer: String::from("SomeManufacturer"),
+            recv_type: String::from("SomeKind"), 
+            serial_number: String::from("XXXXXX"), 
+            year: 2021, 
+            software_number: String::from("v00"),
+        };
+        cggtts.set_rcvr_infos(rcvr);
+
+        // add some more infos
+        cggtts.set_lab_agency("MyLab");
+        cggtts.set_nb_channels(10);
+        cggtts.set_antenna_coordinates((1.0,2.0,3.0));
+        cggtts.set_time_reference("UTC(k)");
+
+        // define a total delay
+        let delay = CalibratedDelay {
+            constellation: track::Constellation::GPS,
+            values: vec![25E-9_f64],
+            codes: vec![String::from("C1")], 
+            report: String::from("NA"),
+        };
+        cggtts.set_internal_delay(delay);
+        cggtts.set_cable_delay(100E-9);
+        cggtts.set_ref_delay(50E-9);
+        let total_delay = cggtts.total_delay();
+        assert_eq!(total_delay.values.len(), 1); // single freq
+        assert_eq!(cggtts.total_delay().values[0], 25E-9+25E-9+100E-9); 
+
+        let fp = std::path::Path::new("data/output/GZXXXXDD.DD4");
+        assert_eq!(cggtts.to_file(fp).is_err(), false)
+    }
+    
+    #[test]
+    /// Another test..
+    fn cggtts_with_ionospheric_parameters () {
+        let mut cggtts = Cggtts::default();
+
+        // identify receiver hw
+        let rcvr = Rcvr {
+            manufacturer: String::from("SomeManuf1"),
+            recv_type: String::from("SomeKind1"), 
+            serial_number: String::from("XXXXXX"), 
+            year: 2021, 
+            software_number: String::from("v01"),
+        };
+        cggtts.set_rcvr_infos(rcvr);
+
+        // IMS infos
+        let ims = Rcvr {
+            manufacturer: String::from("SomeManuf2"),
+            recv_type: String::from("SomeKind2"), 
+            serial_number: String::from("YYYY"), 
+            year: 2022,
+            software_number: String::from("v02"),
+        };
+        cggtts.set_ims_infos(ims);
+
+        // add some more infos
+        cggtts.set_lab_agency("MyLab");
+        cggtts.set_nb_channels(10);
+        cggtts.set_antenna_coordinates((1.0,2.0,3.0));
+        cggtts.set_time_reference("UTC(USNO)");
+
+        // define a delay
+        let delay = CalibratedDelay {
+            constellation: track::Constellation::GPS,
+            values: vec![25E-9_f64],
+            codes: vec![String::from("C1")], 
+            report: String::from("NA"),
+        };
+        cggtts.set_internal_delay(delay);
+        cggtts.set_cable_delay(100E-9);
+        cggtts.set_ref_delay(50E-9);
+
+        // add some measurements
+        let mut track = track::CggttsTrack::default();
+        track.set_satellite_id(0x01);
+        cggtts.tracks.push(track);
+        let mut track = track::CggttsTrack::default();
+        track.set_satellite_id(0x11);
+        cggtts.tracks.push(track);
+
+        let total_delay = cggtts.total_delay();
+        assert_eq!(total_delay.values.len(), 1); // single freq
+        assert_eq!(cggtts.total_delay().values[0], 25E-9+25E-9+100E-9); 
+
+        let fp = std::path::Path::new("data/output/GZXXXXDD.DD5");
+        assert_eq!(cggtts.to_file(fp).is_err(), false)
+    }
 }
